from __future__ import annotations

from typing import TYPE_CHECKING

import pytest_asyncio

from picodi import shutdown_dependencies

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator


@pytest_asyncio.fixture()
async def _picodi_shutdown() -> AsyncGenerator[None, None]:
    """
    Shutdown dependencies after the test (async version).
    Need for tests consistency.
    """
    yield
    await shutdown_dependencies()



from __future__ import annotations

from typing import TYPE_CHECKING

import picodi

if TYPE_CHECKING:
    from starlette.types import ASGIApp, Receive, Scope, Send


class RequestScope(picodi.ContextVarScope):
    """
    Request scope. Values cached in contextvars.
    Dependencies initialized on request start and closed on request end.
    """


class RequestScopeMiddleware:
    """
    Starlette Pure ASGI Middleware for automatically
    initializing and closing request scoped dependencies
    """

    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        await picodi.init_dependencies(scope_class=RequestScope)
        try:
            await self.app(scope, receive, send)
        finally:
            await picodi.shutdown_dependencies(  # noqa: ASYNC102
                scope_class=RequestScope
            )



from __future__ import annotations

from typing import Any

from fastapi import Depends as FastAPIDepends

from picodi._picodi import DependencyCallable, Depends
from picodi.integrations.starlette import RequestScope, RequestScopeMiddleware

__all__ = [
    "Provide",
    "RequestScope",
    "RequestScopeMiddleware",
]


class DependsAsyncCallable(Depends):
    async def __call__(self) -> Any:
        return self


def Provide(  # noqa: N802
    dependency: DependencyCallable, /, *, wrap: bool = False
) -> Any:
    """
    Drop-in replacement for :func:`picodi.Provide` but for FastAPI.

    :param dependency: callable dependency.
    :param wrap: wrap dependency in ``fastapi.Depends``.
    """
    dep = DependsAsyncCallable(dependency)
    return FastAPIDepends(dep) if wrap else dep  # type: ignore[arg-type]



from __future__ import annotations

import asyncio
import copy
import functools
import inspect
import logging
import threading
from collections.abc import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    Iterable,
    Iterator,
)
from contextlib import asynccontextmanager, contextmanager
from dataclasses import asdict, dataclass
from typing import (
    Annotated,
    Any,
    ContextManager,
    NamedTuple,
    ParamSpec,
    TypeVar,
    cast,
    get_origin,
    overload,
)

from picodi._scopes import (
    AutoScope,
    ContextVarScope,
    ManualScope,
    NullScope,
    ScopeType,
    SingletonScope,
)
from picodi.support import ExitStack, NullAwaitable

try:
    import fastapi.params
except ImportError:  # pragma: no cover
    fastapi = None  # type: ignore[assignment] # pragma: no cover


logger = logging.getLogger("picodi")


DependencyCallable = Callable[..., Any]
LifespanScopeClass = type[ManualScope] | tuple[type[ManualScope], ...]
T = TypeVar("T")
P = ParamSpec("P")
TC = TypeVar("TC", bound=Callable)

unset = object()


class RegistryStorage:
    def __init__(self) -> None:
        self.deps: dict[DependencyCallable, Provider] = {}
        self.overrides: dict[DependencyCallable, DependencyCallable] = {}
        self.touched_dependencies: set[DependencyCallable] = set()
        self.lock = threading.RLock()

    def __iter__(self) -> Iterator[Provider]:
        return iter(self.deps.values())


class InternalRegistry:
    def __init__(self, storage: RegistryStorage) -> None:
        self._storage = storage

    def add(
        self,
        dependency: DependencyCallable,
        scope_class: type[ScopeType] = NullScope,
        override_scope: bool = False,
        ignore_manual_init: bool | Callable[[], bool] = False,
    ) -> None:
        """
        Add a dependency to the registry.
        """
        with self._storage.lock:
            if dependency in self._storage.deps:
                provider = self._storage.deps[dependency]
                to_replace = provider.replace(
                    scope_class=(scope_class if override_scope else None)
                )
                if to_replace != provider:
                    self._storage.deps[dependency] = to_replace
            else:
                self._storage.deps[dependency] = Provider.from_dependency(
                    dependency=dependency,
                    scope_class=scope_class,
                    ignore_manual_init=ignore_manual_init,
                )

    def get(self, dependency: DependencyCallable) -> Provider:
        with self._storage.lock:
            dependency = self.get_dep_or_override(dependency)
            self._storage.touched_dependencies.add(dependency)
            return self._storage.deps[dependency]

    def get_dep_or_override(self, dependency: DependencyCallable) -> DependencyCallable:
        return self._storage.overrides.get(dependency, dependency)

    def get_override(self, dependency: DependencyCallable) -> DependencyCallable | None:
        return self._storage.overrides.get(dependency)

    def get_original(self, override: DependencyCallable) -> DependencyCallable | None:
        for original, overriden in self._storage.overrides.items():
            if overriden == override:
                return original
        return None

    def has_overrides(self) -> bool:
        return bool(self._storage.overrides)

    def filter(self, predicate: Callable[[Provider], bool]) -> Iterable[Provider]:
        return filter(predicate, self._storage)


class Registry:
    """
    Manages dependencies and overrides.
    """

    def __init__(
        self, storage: RegistryStorage, internal_registry: InternalRegistry
    ) -> None:
        self._storage = storage
        self._internal_registry = internal_registry

    @property
    def touched(self) -> frozenset[DependencyCallable]:
        """
        Get all dependencies that were used during the picodi lifecycle.
        This method will return a frozenset of dependencies that were resolved.
        It will not include dependencies that were overridden.
        Primarily used for testing purposes.
        For example, you can check that mongo
        database was used in the test and clear it after the test.
        """
        return frozenset(self._storage.touched_dependencies)

    @overload
    def override(
        self, dependency: DependencyCallable, new_dependency: None = None
    ) -> Callable[[DependencyCallable], DependencyCallable]: ...

    @overload
    def override(
        self, dependency: DependencyCallable, new_dependency: DependencyCallable
    ) -> ContextManager[None]: ...

    def override(
        self,
        dependency: DependencyCallable,
        new_dependency: DependencyCallable | None | object = unset,
    ) -> Callable[[DependencyCallable], DependencyCallable] | ContextManager[None]:
        """
        Override a dependency with a new one. It can be used as a decorator,
        as a context manager or as a regular method call. New dependency will be
        added to the registry.

        :param dependency: dependency to override
        :param new_dependency: new dependency to use. If explicitly set to ``None``,
            it will remove the override.

        Examples
        --------
        .. code-block:: python

            @registry.override(get_settings)
            def real_settings():
                return {"real": "settings"}


            with registry.override(get_settings, real_settings):
                ...

            registry.override(get_settings, real_settings)
            registry.override(get_settings, None)  # clear override
        """
        if _internal_registry.get_original(dependency):
            raise ValueError("Cannot override an overridden dependency")

        def decorator(override_to: DependencyCallable) -> DependencyCallable:
            self._internal_registry.add(override_to)
            if dependency is override_to:
                raise ValueError("Cannot override a dependency with itself")
            self._storage.overrides[dependency] = override_to
            return override_to

        if new_dependency is unset:
            return decorator

        with self._storage.lock:
            call_dependency = self._storage.overrides.get(dependency)
            if callable(new_dependency):
                decorator(new_dependency)
            else:
                self._storage.overrides.pop(dependency, None)

        @contextmanager
        def manage_context() -> Generator[None, None, None]:
            try:
                yield
            finally:
                self.override(dependency, call_dependency)

        return manage_context()

    def clear_overrides(self) -> None:
        """
        Clear all overrides. It will remove all overrides, but keep the dependencies.
        """
        with self._storage.lock:
            self._storage.overrides.clear()

    def clear_touched(self) -> None:
        """
        Clear the touched dependencies.
        It will remove all dependencies resolved during the picodi lifecycle.
        """
        with self._storage.lock:
            self._storage.touched_dependencies.clear()

    def clear(self) -> None:
        """
        Clear the registry. It will remove all dependencies and overrides.
        This method will not close any dependencies. So you need to manually call
        :func:`shutdown_dependencies` before this method.
        """
        with self._storage.lock:
            self._storage.deps.clear()
            self._storage.overrides.clear()
            self._storage.touched_dependencies.clear()


_registry_storage = RegistryStorage()
_internal_registry = InternalRegistry(_registry_storage)
registry = Registry(_registry_storage, _internal_registry)
_scopes: dict[type[ScopeType], ScopeType] = {
    NullScope: NullScope(),
    SingletonScope: SingletonScope(),
    ContextVarScope: ContextVarScope(),
}
_lock = threading.RLock()


def Provide(dependency: DependencyCallable, /) -> Any:  # noqa: N802
    """
    Declare a provider.
    It takes a single "dependency" callable (like a function).
    Don't call it directly, picodi will call it for you.

    :param dependency: can be a regular function or a generator with one yield.
        If the dependency is a generator, it will be used as a context manager.
        Any generator that is valid for :func:`python:contextlib.contextmanager`
        can be used as a dependency.

    Example
    -------
    .. code-block:: python

        from picodi import Provide, inject


        def get_db():
            yield "db connection"
            print("closing db connection")


        @inject
        def my_service(db: str = Provide(get_db)):
            assert db == "db connection"
    """
    return Depends(dependency)


def inject(fn: Callable[P, T]) -> Callable[P, T]:
    """
    Decorator to inject dependencies into a function.
    Use it in combination with :func:`Provide` to declare dependencies.
    Should be placed first in the decorator chain (on bottom).

    :param fn: function to decorate.

    Example
    -------
    .. code-block:: python

        from picodi import inject, Provide


        @inject
        def my_service(db=Provide(some_dependency_func)): ...
    """
    signature = inspect.signature(fn)
    dependant = _build_depend_tree(Depends(fn))

    if inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn):

        @functools.wraps(fn)
        async def fun_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            gen = _wrapper_helper(
                dependant,
                signature,
                is_async=True,
                args=args,
                kwargs=kwargs,
            )
            value, action = next(gen)
            result = None
            exceptions = []
            while True:
                if inspect.iscoroutine(value):
                    try:
                        value = await value
                    except Exception as e:  # noqa: PIE786
                        exceptions.append(e)

                if action == "result":
                    result = value
                try:
                    value, action = gen.send(value)
                except StopIteration:
                    break
            if exceptions:
                # TODO use `ExceptionGroup` after dropping 3.10 support
                raise exceptions[0]
            return cast("T", result)

        wrapper = fun_wrapper

        if inspect.isasyncgenfunction(fn):

            @functools.wraps(fn)
            async def gen_wrapper(
                *args: P.args, **kwargs: P.kwargs
            ) -> AsyncGenerator[T, None]:
                result = await fun_wrapper(*args, **kwargs)
                async for value in result:  # type: ignore[attr-defined]
                    try:
                        yield value
                    except Exception as e:  # noqa: PIE786
                        try:
                            await result.athrow(e)  # type: ignore[attr-defined]
                        except StopAsyncIteration:
                            break

            wrapper = gen_wrapper  # type: ignore[assignment]

    else:

        @functools.wraps(fn)
        def fun_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            gen = _wrapper_helper(
                dependant,
                signature,
                is_async=False,
                args=args,
                kwargs=kwargs,
            )
            value, action = next(gen)
            result = None
            while True:
                if action == "result":
                    result = value
                try:
                    value, action = gen.send(value)
                except StopIteration:
                    break
            return cast("T", result)

        wrapper = fun_wrapper

        if inspect.isgeneratorfunction(fn):

            @functools.wraps(fn)
            def gen_wrapper(
                *args: P.args, **kwargs: P.kwargs
            ) -> Generator[T, None, None]:
                yield from fun_wrapper(*args, **kwargs)  # type: ignore[misc]

            wrapper = gen_wrapper  # type: ignore[assignment]

    return wrapper  # type: ignore[return-value]


def dependency(
    *,
    scope_class: type[ScopeType] = NullScope,
    ignore_manual_init: bool | Callable[[], bool] = False,
) -> Callable[[TC], TC]:
    """
    Decorator to declare a dependency. You don't need to use it with default arguments,
    use it only if you want to change the scope of the dependency.
    Should be placed last in the decorator chain (on top).

    :param scope_class: specify the scope class to use for the dependency. Default is
        :class:`NullScope`.
        Picodi additionally provides a few built-in scopes:
        :class:`SingletonScope`, :class:`ContextVarScope`.
    :param ignore_manual_init: this parameter can be used to skip the dependency manual
        initialization. It can be a boolean or a callable that returns a boolean.
        If it's a callable, it will be called every time before the dependency is
        initialized with :func:`init_dependencies`. If it returns True, the dependency
        will be skipped.
    """

    if scope_class not in _scopes:
        _scopes[scope_class] = scope_class()

    def decorator(fn: TC) -> TC:
        _internal_registry.add(
            fn,
            scope_class=scope_class,
            override_scope=True,
            ignore_manual_init=ignore_manual_init,
        )
        return fn

    return decorator


def init_dependencies(scope_class: LifespanScopeClass = SingletonScope) -> Awaitable:
    """
    Call this function to close dependencies. Usually, it should be called
    when your application is starting up.

    This function works both for synchronous and asynchronous dependencies.
    If you call it without ``await``, it will initialize only sync dependencies.
    If you call it ``await init_dependencies()``, it will initialize both sync and async
    dependencies.

    If you not pass any arguments, it will initialize only :class:`SingletonScope`
    and its subclasses.

    :param scope_class: you can specify the scope class to initialize. If passed -
        only dependencies of this scope class and its subclasses will be initialized.
    """
    async_deps = []
    filtered_providers = _internal_registry.filter(
        lambda p: not p.is_ignored() and issubclass(p.scope_class, scope_class)
    )
    for provider in filtered_providers:
        resolver = LazyResolver(provider)
        value = resolver(provider.is_async)
        if provider.is_async:
            async_deps.append(value)

    if async_deps:
        # asyncio.gather runs coros in different tasks with different context
        #   we can run them in current context with contextvars.copy_context()
        #   but in this case `ContextVarScope` will save values in the wrong context.
        #   So we fix it in dirty way by running all coros one by one until
        #   come up with better solution.
        async def init_all() -> None:
            for dep in async_deps:
                await dep

        return init_all()
    return NullAwaitable()


def shutdown_dependencies(
    scope_class: LifespanScopeClass = ManualScope,
) -> Awaitable:
    """
    Call this function to close dependencies. Usually, it should be called
    when your application is shut down.

    This function works both for synchronous and asynchronous dependencies.
    If you call it without ``await``, it will shutdown only sync dependencies.
    If you call it ``await shutdown_dependencies()``, it will shutdown both
    sync and async dependencies.

    If you not pass any arguments, it will shutdown subclasses of :class:`ManualScope`.

    :param scope_class: you can specify the scope class to shutdown. If passed -
        only dependencies of this scope class and its subclasses will be shutdown.
    """
    tasks = [
        instance.shutdown()  # type: ignore[call-arg]
        for klass, instance in _scopes.items()
        if issubclass(klass, scope_class)
    ]
    if all(isinstance(task, NullAwaitable) for task in tasks):
        return NullAwaitable()
    return asyncio.gather(*tasks)


class Depends(NamedTuple):
    call: DependencyCallable


@dataclass(frozen=True)
class Provider:
    dependency: DependencyCallable
    is_async: bool
    scope_class: type[ScopeType]
    ignore_manual_init: bool | Callable[[], bool]

    @classmethod
    def from_dependency(
        cls,
        dependency: DependencyCallable,
        scope_class: type[ScopeType],
        ignore_manual_init: bool | Callable[[], bool] = False,
    ) -> Provider:
        is_async = inspect.iscoroutinefunction(
            dependency
        ) or inspect.isasyncgenfunction(dependency)
        return cls(
            dependency=dependency,
            is_async=is_async,
            scope_class=scope_class,
            ignore_manual_init=ignore_manual_init,
        )

    def is_ignored(self) -> bool:
        if callable(self.ignore_manual_init):
            return self.ignore_manual_init()
        return self.ignore_manual_init

    def replace(self, scope_class: type[ScopeType] | None = None) -> Provider:
        kwargs = asdict(self)
        if scope_class is not None:
            kwargs["scope_class"] = scope_class
        return Provider(**kwargs)

    def get_scope(self) -> ScopeType:
        return _scopes[self.scope_class]

    def resolve_value(self, exit_stack: ExitStack | None, **kwargs: Any) -> Any:
        scope = self.get_scope()
        value_or_gen = self.dependency(**kwargs)
        if self.is_async:

            async def resolve_value_inner() -> Any:
                value_or_gen_ = value_or_gen
                if inspect.iscoroutine(value_or_gen):
                    value_or_gen_ = await value_or_gen_
                if inspect.isasyncgen(value_or_gen_):
                    context_manager = asynccontextmanager(
                        lambda *args, **kwargs: value_or_gen_
                    )
                    if isinstance(scope, AutoScope):
                        assert exit_stack is not None, "exit_stack is required"
                        return await scope.enter(exit_stack, context_manager())
                    return await scope.enter(context_manager())
                return value_or_gen_

            return resolve_value_inner()

        if inspect.isgenerator(value_or_gen):
            context_manager = contextmanager(lambda *args, **kwargs: value_or_gen)
            if isinstance(scope, AutoScope):
                assert exit_stack is not None, "exit_stack is required"
                return scope.enter(exit_stack, context_manager())
            return scope.enter(context_manager())
        return value_or_gen


def _wrapper_helper(
    dependant: DependNode,
    signature: inspect.Signature,
    is_async: bool,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Generator[Any, None, None]:
    exit_stack = ExitStack()
    if _need_patch(dependant):
        dependant = copy.deepcopy(dependant)
        _patch_dependant(dependant)
    bound = signature.bind(*args, **kwargs)
    bound.apply_defaults()
    arguments: dict[str, Any] = bound.arguments
    scopes: list[ScopeType] = []
    is_root = any(isinstance(value, Depends) for value in bound.arguments.values())

    if is_root:
        arguments, scopes = _resolve_dependencies(dependant, exit_stack)

    for scope in scopes:
        scope.enter_inject()
    for name, call in arguments.items():
        if isinstance(call, LazyResolver):
            value = yield call(is_async=is_async), "dependency"
            bound.arguments[name] = value

    try:
        result = dependant.value.call(*bound.args, **bound.kwargs)
    except Exception as e:
        for scope in scopes:
            scope.exit_inject(e)
            if isinstance(scope, AutoScope):
                yield scope.shutdown(exit_stack, e), "close_scope"
        raise

    if inspect.isgenerator(result):

        @functools.wraps(result)  # type: ignore[arg-type]
        def gen() -> Generator[Any, None, None]:
            exception = None
            try:
                yield from result
            except Exception as e:
                exception = e
                raise
            finally:
                for scope in scopes:
                    scope.exit_inject(exception)
                    if isinstance(scope, AutoScope):
                        scope.shutdown(exit_stack, exception)

        yield gen(), "result"
        return
    elif inspect.isasyncgen(result):

        @functools.wraps(result)  # type: ignore[arg-type]
        async def gen() -> AsyncGenerator[Any, None]:
            exception = None
            try:
                async for item in result:
                    yield item
            except Exception as e:
                exception = e
                raise
            # TODO use
            #   https://flake8-async.readthedocs.io/en/latest/glossary.html#cancel-scope
            #   https://docs.python.org/3/library/asyncio-task.html#task-cancellation
            finally:
                for scope in scopes:
                    scope.exit_inject(exception)
                    if isinstance(scope, AutoScope):
                        await scope.shutdown(exit_stack, exception)  # noqa: ASYNC102

        yield gen(), "result"
        return

    yield result, "result"
    for scope in scopes:
        scope.exit_inject()
        if isinstance(scope, AutoScope):
            yield scope.shutdown(exit_stack), "close_scope"


def _need_patch(dependant: DependNode) -> bool:
    if dependant.name is None and not _internal_registry.has_overrides():
        return False

    for dep in dependant.dependencies:  # noqa: SIM110
        if _need_patch(dep):
            return True

    return bool(
        dependant.name and _internal_registry.get_override(dependant.value.call)
    )


def _patch_dependant(dependant: DependNode) -> None:
    for dep in dependant.dependencies:
        _patch_dependant(dep)

    if dependant.name is None:
        return

    if override := _internal_registry.get_override(dependant.value.call):
        dependant.value = Depends(override)
        override_tree = _build_depend_tree(dependant.value)
        dependant.dependencies = override_tree.dependencies


def _resolve_dependencies(
    dependant: DependNode, exit_stack: ExitStack
) -> tuple[dict[str, LazyResolver], list[ScopeType]]:
    scopes = set()
    resolved_dependencies = {}
    for dep in dependant.dependencies:
        values, dep_scopes = _resolve_dependencies(dep, exit_stack)
        resolved_dependencies.update(values)
        scopes.update(dep_scopes)

    if dependant.name is None:
        return resolved_dependencies, list(scopes)

    provider = _internal_registry.get(dependant.value.call)
    value = LazyResolver(
        provider=provider,
        kwargs=resolved_dependencies,
        exit_stack=exit_stack,
    )
    return {dependant.name: value}, [provider.get_scope()]


@dataclass
class DependNode:
    value: Depends
    name: str | None
    dependencies: list[DependNode]


def _build_depend_tree(dependency: Depends, name: str | None = None) -> DependNode:
    signature = inspect.signature(dependency.call)
    dependencies = []
    for name_, value in signature.parameters.items():
        param_dep = _extract_and_register_dependency_from_parameter(value)
        if param_dep is not None:
            dependencies.append(_build_depend_tree(param_dep, name=name_))
    return DependNode(value=dependency, dependencies=dependencies, name=name)


def _extract_and_register_dependency_from_parameter(
    value: inspect.Parameter,
) -> Depends | None:
    if isinstance(value.default, Depends):
        _internal_registry.add(value.default.call)
        return value.default

    if fastapi is None:
        return None  # type: ignore[unreachable]  # pragma: no cover
    fastapi_dependency = None
    if isinstance(value.default, fastapi.params.Depends):
        fastapi_dependency = value.default.dependency
    elif get_origin(value.annotation) is Annotated:
        for metadata in value.annotation.__metadata__:
            if isinstance(metadata, fastapi.params.Depends):
                fastapi_dependency = metadata.dependency
                break
    if isinstance(fastapi_dependency, Depends):
        _internal_registry.add(fastapi_dependency.call)  # type: ignore[unreachable]
        return fastapi_dependency
    return None


class LazyResolver:
    def __init__(
        self,
        provider: Provider,
        kwargs: dict[str, Any] | None = None,
        exit_stack: ExitStack | None = None,
    ) -> None:
        self.provider = provider
        self.kwargs = kwargs or {}
        self.exit_stack = exit_stack

    def __call__(self, is_async: bool) -> Any:
        call = self._resolve_async if is_async else self._resolve
        return call()

    def _resolve(self) -> Any:
        scope = self.provider.get_scope()
        try:
            value = scope.get(self.provider.dependency)
        except KeyError:
            if self.provider.is_async:
                value = self.provider.dependency()
            else:
                with _lock:
                    try:
                        value = scope.get(self.provider.dependency)
                    except KeyError:
                        value = self.provider.resolve_value(
                            self.exit_stack, **self.kwargs
                        )
                        scope.set(self.provider.dependency, value)
        return value

    async def _resolve_async(self) -> Any:
        scope = self.provider.get_scope()
        try:
            value = scope.get(self.provider.dependency)
        except KeyError:
            with _lock:
                try:
                    value = scope.get(self.provider.dependency)
                except KeyError:
                    value = self.provider.resolve_value(self.exit_stack, **self.kwargs)
                    if self.provider.is_async:
                        value = await value
                    scope.set(self.provider.dependency, value)
        return value



"""
Helper functions and classes for picodi.
"""

from __future__ import annotations

import asyncio
import contextlib
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncContextManager,
    ContextManager,
    Generic,
    ParamSpec,
    TypeVar,
    overload,
)

from picodi import (
    ManualScope,
    Provide,
    SingletonScope,
    init_dependencies,
    inject,
    shutdown_dependencies,
)

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Coroutine, Generator

    from picodi._picodi import LifespanScopeClass

sentinel = object()


class PathNotFoundError(Exception):
    def __init__(self, current_path: str, obj: Any):
        self.current_path = current_path
        self.obj = obj
        super().__init__(f"Can't find path '{current_path}' in {type(obj)} object")


def get_value(path: str, obj: Any, *, default: Any = sentinel) -> Any:
    """
    Get attribute from nested objects.
    Can be useful to avoid passing entire objects like
    settings to functions that only need a small part of it.

    :param path: path to the attribute, separated by dots.
    :param obj: object to search the attribute in.
    :param default: default value to return if the attribute is not found.
        If not provided, an :exc:`PathNotFoundError` is raised.

    :raises PathNotFoundError: if the path is not found in the object
        and default is not provided.

    Example
    -------
    .. code-block:: python

        obj = SimpleNamespace(foo=SimpleNamespace(bar={"baz": 42}))

        get_value("foo.bar.baz", obj)  # Output: 42
        get_value("foo.bar.baz2", obj)  # Output: AttributeError
        get_value("foo.bar.baz2", obj, default=12)  # Output: 12

    """
    if not path:
        raise ValueError("Empty path")
    if not isinstance(path, str):
        raise TypeError("Path must be a string")

    value = obj
    current_path = []
    for part in path.split("."):
        current_path.append(part)
        curr_val = _get_attr(value, part)
        if curr_val is sentinel:
            curr_val = _get_item(value, part)
        if curr_val is sentinel:
            if default is sentinel:
                raise PathNotFoundError(".".join(current_path), obj)
            return default
        value = curr_val

    return value


def _get_attr(obj: Any, attr: str) -> Any:
    try:
        return getattr(obj, attr)
    except AttributeError:
        return sentinel


def _get_item(obj: Any, key: str) -> Any:
    try:
        return obj[key]
    except (KeyError, TypeError):
        return sentinel


T = TypeVar("T")
P = ParamSpec("P")


class _Lifespan:
    """
    Lifecycle manager for dependencies.

    Example
    -------
    .. code-block:: python

        from picodi.helpers import lifespan


        @lifespan
        def main(): ...


        @lifespan
        async def async_main(): ...


        @lifespan.sync()
        def main(): ...


        @lifespan.async_()
        async def async_main(): ...


        with lifespan.sync():
            ...

        async with lifespan.async_():
            ...
    """

    @overload
    def __call__(self, fn: Callable[P, T]) -> Callable[P, T]:
        """
        Decorator for functions

        :param fn: function to decorate.
        """

    @overload
    def __call__(
        self,
        fn: None = None,
        *,
        init_scope_class: LifespanScopeClass | None = SingletonScope,
        shutdown_scope_class: LifespanScopeClass | None = ManualScope,
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        """Sync and Async context manager"""

    def __call__(
        self,
        fn: Callable[P, T] | None = None,
        *,
        init_scope_class: LifespanScopeClass | None = SingletonScope,
        shutdown_scope_class: LifespanScopeClass | None = ManualScope,
    ) -> Callable[P, T] | Callable[[Callable[P, T]], Callable[P, T]]:
        """
        Can be used as a decorator or a context manager.

        :param fn: function to decorate (if used as a decorator).
        :param init_scope_class: scope class for initialization
            (can be omitted by passing None).
        :param shutdown_scope_class: scope class for shutdown
            (can be omitted by passing None).
        :return: decorated function or decorator.
        """

        def decorator(fn: Callable[P, T]) -> Callable[P, T]:
            if asyncio.iscoroutinefunction(fn):
                return self.async_(  # type: ignore[return-value]
                    init_scope_class=init_scope_class,
                    shutdown_scope_class=shutdown_scope_class,
                )(fn)
            return self.sync(
                init_scope_class=init_scope_class,
                shutdown_scope_class=shutdown_scope_class,
            )(fn)

        return decorator if fn is None else decorator(fn)

    @contextlib.contextmanager
    def sync(
        self,
        *,
        init_scope_class: LifespanScopeClass | None = SingletonScope,
        shutdown_scope_class: LifespanScopeClass | None = ManualScope,
    ) -> Generator[None, None, None]:
        """
        :attr:`lifespan` can automatically detect if the decorated function
        is async or not. But if you want to force sync behavior, ``lifespan.sync``.

        :param init_scope_class: scope class for initialization
            (can be omitted by passing None).
        :param shutdown_scope_class: scope class for shutdown
            (can be omitted by passing None).
        """
        if init_scope_class is not None:
            init_dependencies(init_scope_class)
        try:
            yield
        finally:
            if shutdown_scope_class is not None:
                shutdown_dependencies(shutdown_scope_class)

    @contextlib.asynccontextmanager
    async def async_(
        self,
        *,
        init_scope_class: LifespanScopeClass | None = SingletonScope,
        shutdown_scope_class: LifespanScopeClass | None = ManualScope,
    ) -> AsyncGenerator[None, None]:
        """
        :attr:`lifespan` can automatically detect if the decorated function
        is async or not.
        But if you want to force async behavior, ``lifespan.async_``.

        :param init_scope_class: scope class for initialization
            (can be omitted by passing None).
        :param shutdown_scope_class: scope class for shutdown
            (can be omitted by passing None).
        """
        if init_scope_class is not None:
            await init_dependencies(init_scope_class)
        try:
            yield
        finally:
            if shutdown_scope_class is not None:
                await shutdown_dependencies(shutdown_scope_class)  # noqa: ASYNC102


lifespan = _Lifespan()


class enter(Generic[T]):  # noqa: N801
    """
    Create a context manager from a dependency.

    Can be handy for testing or when you need to create "fabric" dependency
    (dependency that creates other dependencies based on some conditions).

    :param dependency: dependency to create a context manager from.
        Like with :func:`Provide` - don't call the dependency function here,
        just pass it.
    :return: sync or async context manager.

    Example
    -------

    .. code-block:: python

        from picodi.helpers import enter


        def get_42():
            yield 42


        with enter(get_42) as val:
            assert val == 42
    """

    def __init__(
        self,
        dependency: Callable[
            [],
            Coroutine[T, None, None]
            | Generator[T, None, None]
            | AsyncGenerator[T, None]
            | T,
        ],
    ) -> None:
        self.dependency = dependency
        self.sync_cm: ContextManager[T] = self._create_sync_cm()
        self.async_cm: AsyncContextManager[T] = self._create_async_cm()

    def _create_sync_cm(self) -> ContextManager[T]:
        @contextlib.contextmanager
        @inject
        def sync_enter(dep: Any = Provide(self.dependency)) -> Generator[T, None, None]:
            yield dep

        return sync_enter()

    def _create_async_cm(self) -> AsyncContextManager[T]:
        @contextlib.asynccontextmanager
        @inject
        async def async_enter(
            dep: Any = Provide(self.dependency),
        ) -> AsyncGenerator[T, None]:
            yield dep

        return async_enter()

    def __enter__(self) -> T:
        return self.sync_cm.__enter__()

    def __exit__(self, *args: Any) -> None:
        self.sync_cm.__exit__(*args)

    async def __aenter__(self) -> T:
        return await self.async_cm.__aenter__()

    async def __aexit__(self, *args: Any) -> None:
        await self.async_cm.__aexit__(*args)



from __future__ import annotations

from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, AsyncContextManager, ContextManager, TypeAlias

from picodi.support import ExitStack, NullAwaitable

if TYPE_CHECKING:
    from collections.abc import Awaitable, Hashable


unset = object()


class Scope:
    """
    Scopes are used to store and retrieve values by key and for closing dependencies.

    Don't use this class directly and don't inherit from it.
    Inherit from :class:`AutoScope` or :class:`ManualScope`.
    """

    def get(self, key: Hashable) -> Any:
        """
        Get a value by key.
        If value is not exists must raise KeyError.

        :param key: key to get value, typically a dependency function.
        :raises KeyError: if value not exists.
        """
        raise NotImplementedError

    def set(self, key: Hashable, value: Any) -> None:
        """
        Set a value by key.

        :param key: key to set value, typically a dependency function.
        :param value: value to set, typically a dependency instance.
        """
        raise NotImplementedError

    def enter_inject(self) -> None:
        """
        Called when entering an :func:`inject` decorator.
        """
        return None

    def exit_inject(self, exc: BaseException | None = None) -> None:  # noqa: U100
        """
        Called before exiting a :func:`inject` decorator.

        ``shutdown`` will be called after this, e.g.:
            ``exit_inject`` -> ``shutdown`` -> ``inject`` wrapper returns.
        """
        return None


class AutoScope(Scope):
    """
    AutoScope is a scope that automatically closes dependencies
    after exiting the context.

    Don't use this class directly.
    Usually, you don't want to inherit this class, but you can.
    """

    def enter(
        self,
        exit_stack: ExitStack,
        context_manager: AsyncContextManager | ContextManager,
    ) -> Awaitable:
        """
        Hook for entering yielded dependencies context. Will be called automatically
        by picodi.
        Usually you don't need to override this method.

        :param exit_stack: :class:`support.ExitStack` instance. Instance of ExitStack
            will be created automatically in :func:`inject` decorator.
        :param context_manager: context manager created from yield dependency.
        """
        return exit_stack.enter_context(context_manager)

    def shutdown(
        self, exit_stack: ExitStack, exc: BaseException | None = None
    ) -> Awaitable:
        """
        Hook for closing dependencies. Will be called automatically by picodi.
        Usually you don't need to override this method.

        :param exit_stack: :class:`support.ExitStack` instance. Instance of ExitStack
            will be created automatically in :func:`inject` decorator.
        :param exc: exception that was raised in the context.
        """
        return exit_stack.close(exc)


class ManualScope(Scope):
    """
    ManualScope is a scope that requires manual closing of dependencies.
    For example :class:`SingletonScope` or :class:`ContextVarScope` use this scope.
    You can close dependencies by calling :func:`shutdown_dependencies` or
    ``shutdown_dependencies(scope_class=MyCustomScope)``
    for shutdown only dependencies that uses :class:`MyCustomScope` scope.

    Don't use this class directly.
    Inherit this class for your custom scope.
    """

    def enter(
        self, context_manager: AsyncContextManager | ContextManager  # noqa: U100
    ) -> Awaitable:
        """
        Hook for entering yielded dependencies context. Will be called automatically
        by picodi or when you call :func:`init_dependencies`.

        :param context_manager: context manager created from yield dependency.
        """
        return NullAwaitable()

    def shutdown(self, exc: BaseException | None = None) -> Awaitable:  # noqa: U100
        """
        Hook for shutdown dependencies.
        Will be called when you call :func:`shutdown_dependencies`

        :param exc: exception that was raised in the context.
        """
        return NullAwaitable()


ScopeType: TypeAlias = AutoScope | ManualScope


class NullScope(AutoScope):
    """
    Null scope.
    Values aren't cached, dependencies closed automatically after function call.
    This is the default scope.
    """

    def get(self, key: Hashable) -> Any:
        raise KeyError(key)

    def set(self, key: Hashable, value: Any) -> None:  # noqa: U100
        return None


class SingletonScope(ManualScope):
    """
    Singleton scope. Values cached for the lifetime of the application.
    Dependencies closed only when user manually call :func:`shutdown_dependencies`.
    """

    def __init__(self) -> None:
        self._exit_stack = ExitStack()
        self._store: dict[Hashable, Any] = {}

    def get(self, key: Hashable) -> Any:
        return self._store[key]

    def set(self, key: Hashable, value: Any) -> None:
        self._store[key] = value

    def enter(self, context_manager: AsyncContextManager | ContextManager) -> Awaitable:
        return self._exit_stack.enter_context(context_manager)

    def shutdown(self, exc: BaseException | None = None) -> Awaitable:
        self._store.clear()
        return self._exit_stack.close(exc)


class ContextVarScope(ManualScope):
    """
    ContextVar scope. Values cached in contextvars.
    Dependencies closed only when user manually call :func:`shutdown_dependencies`.
    """

    def __init__(self) -> None:
        self._exit_stack: ContextVar[ExitStack] = ContextVar(
            "picodi_ContextVarScope_exit_stack"
        )
        self._store: dict[Any, ContextVar[Any]] = {}

    def get(self, key: Hashable) -> Any:
        try:
            value = self._store[key].get()
        except LookupError:
            raise KeyError(key) from None
        if value is unset:
            raise KeyError(key)
        return value

    def set(self, key: Hashable, value: Any) -> None:
        try:
            var = self._store[key]
        except KeyError:
            var = self._store[key] = ContextVar("picodi_ContextVarScope_var")
        var.set(value)

    def enter(self, context_manager: AsyncContextManager | ContextManager) -> Awaitable:
        exit_stack = self._get_exit_stack()
        return exit_stack.enter_context(context_manager)

    def shutdown(self, exc: BaseException | None = None) -> Any:
        for var in self._store.values():
            var.set(unset)
        exit_stack = self._get_exit_stack()
        return exit_stack.close(exc)

    def _get_exit_stack(self) -> ExitStack:
        try:
            stack = self._exit_stack.get()
        except LookupError:
            stack = ExitStack()
            self._exit_stack.set(stack)
        return stack



from picodi._picodi import (
    Provide,
    dependency,
    init_dependencies,
    inject,
    registry,
    shutdown_dependencies,
)
from picodi._scopes import (
    AutoScope,
    ContextVarScope,
    ManualScope,
    NullScope,
    Scope,
    ScopeType,
    SingletonScope,
)

__all__ = [
    "inject",
    "Provide",
    "dependency",
    "init_dependencies",
    "shutdown_dependencies",
    "NullScope",
    "SingletonScope",
    "ContextVarScope",
    "ScopeType",
    "Scope",
    "ManualScope",
    "AutoScope",
    "registry",
]



from __future__ import annotations

from contextlib import ExitStack
from typing import TYPE_CHECKING

import pytest

from picodi import registry, shutdown_dependencies

if TYPE_CHECKING:
    from collections.abc import Callable, Generator


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        "markers", "picodi_override(dep, override): Picodi deps override"
    )


@pytest.fixture()
def _picodi_shutdown() -> Generator[None, None, None]:
    """
    Shutdown dependencies after the test.
    Need for tests consistency.
    """
    yield
    shutdown_dependencies()


@pytest.fixture()
def _picodi_clear_touched() -> Generator[None, None, None]:
    """
    Clear touched dependencies after the test.
    Need for tests consistency.
    """
    yield
    registry.clear_touched()


@pytest.fixture(autouse=True)
def _picodi_teardown(_picodi_clear_touched: None, _picodi_shutdown: None) -> None:
    """
    Automatically cleanup Picodi dependencies and registry after the test.
    """


@pytest.fixture()
def picodi_overrides() -> list[tuple[Callable, Callable]]:
    """
    Get overrides from markers that will be used in the test.

    Usually, you don't need to use this fixture directly.
    Use `picodi_override` marker instead. But if you want to stick to
    some custom logic you can inherit from this fixture.
    """
    return []


@pytest.fixture()
def picodi_overrides_from_marks(
    request: pytest.FixtureRequest,
) -> list[tuple[Callable, Callable]]:
    for marker in request.node.iter_markers(name="picodi_override"):
        if len(marker.args) not in (1, 2):
            pytest.fail(
                "picodi_override marker must have 2 arguments: dependency and override "
                "OR 1 argument: iterable of tuples with dependency and override"
            )

        overrides: list[tuple[Callable, Callable]] = (
            marker.args[0] if len(marker.args) == 1 else [marker.args]
        )
        if not isinstance(overrides, (list, tuple)):
            pytest.fail("Overrides must be a list or tuple")
        for dep, override in overrides:
            if not callable(dep) or not callable(override):
                pytest.fail("Dependency and override must be callable")

        return overrides
    return []


@pytest.fixture(autouse=True)
def _picodi_override_setup(
    picodi_overrides: list[tuple[Callable, Callable]],
    picodi_overrides_from_marks: list[tuple[Callable, Callable]],
) -> Generator[None, None, None]:
    overrides = []
    for item in picodi_overrides_from_marks + picodi_overrides:
        if item not in overrides:
            overrides.append(item)
    with ExitStack() as stack:
        for get_dep, override in overrides:
            stack.enter_context(registry.override(get_dep, override))
        yield
